import { useTranslation } from 'react-i18next'
import { useEffect, useRef, useState } from 'react'
import EntryContainer from './entry-container'
import EntryCallout from './entry-callout'
import EntryActions from './entry-actions'
import AutoExpandingTextArea from '../../../../../shared/components/auto-expanding-text-area'
import AddCommentButton from '../add-comment-button'
import Icon from '../../../../../shared/components/icon'
import {
  useReviewPanelUpdaterFnsContext,
  useReviewPanelValueContext,
} from '../../../context/review-panel/review-panel-context'
import classnames from 'classnames'

function AddCommentEntry() {
  const { t } = useTranslation()
  const { isAddingComment, unsavedComment } = useReviewPanelValueContext()
  const {
    setIsAddingComment,
    submitNewComment,
    handleLayoutChange,
    setUnsavedComment,
  } = useReviewPanelUpdaterFnsContext()

  const [content, setContent] = useState(unsavedComment)
  const [isSubmitting, setIsSubmitting] = useState(false)

  const handleStartNewComment = () => {
    setIsAddingComment(true)
    handleLayoutChange({ async: true })
  }

  const handleSubmitNewComment = async () => {
    setIsSubmitting(true)
    try {
      await submitNewComment(content)
      setIsSubmitting(false)
      setIsAddingComment(false)
      setContent('')
    } catch (err) {
      setIsSubmitting(false)
    }
    handleLayoutChange({ async: true })
  }

  const handleCancelNewComment = () => {
    setIsAddingComment(false)
    setContent('')
    handleLayoutChange({ async: true })
  }

  useEffect(() => {
    return () => {
      setIsAddingComment(false)
    }
  }, [setIsAddingComment])

  const unsavedCommentRef = useRef(unsavedComment)

  // Keep unsaved comment ref up to date for use when the component unmounts
  useEffect(() => {
    unsavedCommentRef.current = content
  }, [content])

  // Store the unsaved comment in the context on unmount
  useEffect(() => {
    return () => {
      setUnsavedComment(unsavedCommentRef.current)
    }
  }, [setUnsavedComment])

  const handleCommentKeyPress = (
    e: React.KeyboardEvent<HTMLTextAreaElement>
  ) => {
    const target = e.target as HTMLTextAreaElement

    if (e.key === 'Enter' && !e.shiftKey && !e.ctrlKey && !e.metaKey) {
      e.preventDefault()
      if (content.length) {
        handleSubmitNewComment()
      }
    }

    if (['PageDown', 'PageUp'].includes(e.key)) {
      if (target.closest('textarea')) {
        e.preventDefault()
      }
    }
  }

  return (
    <EntryContainer id="add-comment">
      <EntryCallout className="rp-entry-callout-add-comment" />
      <div
        className={classnames('rp-entry', 'rp-entry-add-comment', {
          'rp-entry-adding-comment': isAddingComment,
        })}
      >
        {isAddingComment ? (
          <>
            <div className="rp-new-comment">
              {isSubmitting ? (
                <div className="rp-loading">
                  <Icon type="spinner" spin />
                </div>
              ) : (
                <AutoExpandingTextArea
                  className="rp-comment-input"
                  onChange={e => setContent(e.target.value)}
                  onKeyPress={handleCommentKeyPress}
                  onResize={handleLayoutChange}
                  placeholder={t('add_your_comment_here')}
                  value={content}
                  autoFocus // eslint-disable-line jsx-a11y/no-autofocus
                />
              )}
            </div>
            <EntryActions>
              <EntryActions.Button
                className="rp-entry-button-cancel"
                onClick={handleCancelNewComment}
              >
                <Icon type="times" /> {t('cancel')}
              </EntryActions.Button>
              <EntryActions.Button
                onClick={handleSubmitNewComment}
                disabled={isSubmitting || !content.length}
              >
                <Icon type="comment" /> {t('comment')}
              </EntryActions.Button>
            </EntryActions>
          </>
        ) : (
          <AddCommentButton onClick={handleStartNewComment}>
            <Icon type="comment" /> {t('add_comment')}
          </AddCommentButton>
        )}
      </div>
    </EntryContainer>
  )
}

export default AddCommentEntry
